/* recent-manager-ui.vala
 *
 * Copyright (C) 2008-2010 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * The recent files/projects manager
 */
public class Valide.RecentManagerUI : Object
{
  private uint id_file = 0;
  private uint id_project = 0;
  private int max_recents = 5;
  private Gtk.ActionGroup recents_action_group;

  /**
   * This signal is emitted when the recent files is updated
   */
  public signal void update_recent_files ();

  /**
   * This signal is emitted when the recent projects is updated
   */
  public signal void update_recent_projects ();

  /**
   * The ui manager
   */
  public UIManager ui_manager
  {
    get;
    construct;
  }

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    construct;
  }

  /**
   * The project manager
   */
  public ProjectManager projects
  {
    get;
    construct;
  }

  /**
   * The list of the recent files
   */
  private List<Gtk.RecentInfo> _recent_files = null;
  public List<Gtk.RecentInfo> recent_files
  {
    get
    {
      int nb_items = 0;
      List<Gtk.RecentInfo> items;
      Gtk.RecentManager recent_manager;

      this._recent_files = null;
      recent_manager = Gtk.RecentManager.get_default ();
      items = recent_manager.get_items ();
      /* filter */
      foreach (Gtk.RecentInfo l in items)
      {
        if (l.has_group ("valide") && FileUtils.test (l.get_uri (), FileTest.EXISTS)
            && (l.get_uri ().has_suffix (".vala")
                || l.get_uri ().has_suffix (".vapi")
                || l.get_uri ().has_suffix (".gs")))
        {
          this._recent_files.prepend (l);
          nb_items++;
          if (nb_items >= this.max_recents)
          {
            break;
          }
        }
      }
      /* sort */
/*
      this._recent_files = g_list_sort (this._recent_files,
                                        (GCompareFunc) sort_recents_mru);
*/
      return this._recent_files;
    }
  }

  /**
   * The list of the recent projects
   */
  private List<Gtk.RecentInfo> _recent_projects = null;
  public List<Gtk.RecentInfo> recent_projects
  {
    get
    {
      int nb_items = 0;
      List<Gtk.RecentInfo> items;
      Gtk.RecentManager recent_manager;

      this._recent_projects = null;
      /* filter */
      recent_manager = Gtk.RecentManager.get_default ();
      items = recent_manager.get_items ();
      foreach (Gtk.RecentInfo l in items)
      {
        if (l.has_group ("valide") && FileUtils.test (l.get_uri (), FileTest.EXISTS)
            && l.get_uri ().has_suffix (".vide"))
        {
          this._recent_projects.prepend (l);
          nb_items++;
          if (nb_items >= this.max_recents)
          {
            break;
          }
        }
      }
      /* sort */
/*
      this._recent_projects = g_list_sort (this._recent_projects,
                                           (GCompareFunc) sort_recents_mru);
*/
      return this._recent_projects;
    }
  }

  private void recents_menu_activate (Gtk.Action sender)
  {
    string uri;
    Gtk.RecentInfo info;

    info = sender.get_data<Gtk.RecentInfo> ("gtk-recent-info");
    uri = info.get_uri ();

    try
    {
      if (uri.has_suffix ("."+Project.FILE_EXT[0]))
      {
        this.projects.open (uri);
      }
      else
      {
        this.documents.create (uri);
      }
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  private void add (string uri, string mime_type)
  {
    Gtk.RecentData data;

    data = Gtk.RecentData ();
    data.display_name = uri;
    data.description = mime_type;
    data.mime_type = mime_type;
    data.app_name = "Val(a)IDE";
    data.app_exec = "valide";
    string[] groups = new string[2];
    groups[0] = "valide";
    groups[1] = null;
    data.groups = groups;
    Gtk.RecentManager.get_default ().add_full (uri, data);
  }

  private void update (UIManager ui_manager, string[] file_ext, string ui_path,
                       ref uint id)
  {
    List<Gtk.Action> actions;
    List<Gtk.RecentInfo> items;
    Gtk.RecentManager recent_manager;
    List<Gtk.RecentInfo> filtered_items = null;

    if (id != 0)
    {
      ui_manager.remove_ui (id);
    }

    this.recents_action_group = new Gtk.ActionGroup ("RecentActions");
    ui_manager.insert_action_group (this.recents_action_group, 0);

    actions = recents_action_group.list_actions ();
    foreach (Gtk.Action l in actions)
    {
      l.activate.disconnect (this.recents_menu_activate);
       recents_action_group.remove_action (l);
    }

    id = ui_manager.new_merge_id ();

    recent_manager = Gtk.RecentManager.get_default ();
    items = recent_manager.get_items ();

    /* filter */
    foreach (Gtk.RecentInfo l in items)
    {
      if (l.has_group ("valide")
          && FileUtils.test (l.get_uri (), FileTest.EXISTS))
      {
        foreach (string e in file_ext)
        {
          if (l.get_uri ().has_suffix (e))
          {
            filtered_items.prepend (l);
            break;
          }
        }
      }
    }
    /* sort */
/*
    filtered_items = g_list_sort (filtered_items,
                                  (GCompareFunc) sort_recents_mru);
*/
    int i = 0;
    foreach (Gtk.RecentInfo info in filtered_items)
    {
      string uri;
      string tip;
      string label;
      string escaped;
      Gtk.Action action;
      string action_name;

      /* clamp */
      if (i >= max_recents)
      {
        break;
      }
      i++;

      action_name = "recent" + file_ext[0] + "-%d".printf (i);
      escaped = Path.get_basename (info.get_display_name ());

      if (i >= 10)
      {
        label = "%d.  %s".printf (i, escaped);
      }
      else
      {
        label = "_%d.  %s".printf (i, escaped);
      }

      uri = info.get_uri_display ();
      /* Translators: %s is a URI */
      tip = _("Open '%s'").printf (uri);

      action = new Gtk.Action (action_name, label, tip, null);
      action.set_data_full ("gtk-recent-info", info, null);

      action.activate.connect (recents_menu_activate);
      this.recents_action_group.add_action (action);

      ui_manager.add_ui (id, ui_path, action_name, action_name,
                         Gtk.UIManagerItemType.MENUITEM, false);
    }
  }

  /**
   * Create a new Valide.RecentManagerIU
   *
   * @param ui_manager The ui manager
   * @param documents The document manager
   * @param projects The projects manager
   */
  public RecentManagerUI (UIManager ui_manager, DocumentManager documents,
                          ProjectManager projects)
  {
    Object (ui_manager: ui_manager, documents: documents, projects: projects);
  }  

  construct
  {
    this.documents.tab_added.connect ((s, d) => {
      this.add_file (d.path);
    });
    this.projects.project_opened.connect ((s, p) => {
      this.add_project (p.filename);
    });
    this.update_file ();
    this.update_project ();
  }

  /**
   * Add a file
   */
  public void add_file (string uri)
  {
    if (uri.has_suffix (".vala"))
    {
      this.add (uri, "vala");
    }
    if (uri.has_suffix (".vapi"))
    {
      this.add (uri, "vapi");
    }
    else if (uri.has_suffix (".gs"))
    {
      this.add (uri, "genie");
    }
    this.update_file ();
  }

  /**
   * Add a project
   */
  public void add_project (string uri)
  {
    this.add (uri, "vide");
    this.update_project ();
  }

  /**
   * Update the recent files
   */
  public void update_file ()
  {
    string[] ext = {".vala", ".vapi", ".gs"};

    this.update (this.ui_manager, ext, "/menubar/file/file-recents-placeholder",
                 ref this.id_file);
    this.update_recent_files ();
  }

  /**
   * Update the recent projects
   */
  public void update_project ()
  {
    string[] ext = {".vide"};

    this.update (this.ui_manager, ext, "/menubar/project/project-recents-placeholder",
                 ref this.id_project);
    this.update_recent_projects ();
  }
}

