/* source-view.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * A widget for show highlighting text
 */
public class Valide.SourceView : Gtk.SourceView
{
  private GotoDialog goto_dialog = null;

  public signal void search ();
  public signal void find_next ();
  public signal void find_prev ();
  public signal void replace ();

  /**
   * The language
   */
  public string language
  {
    get;
    construct;
  }

  /**
   * The buffer
   */
  public new SourceBuffer buffer
  {
    get
    {
      return base.buffer as SourceBuffer;
    }
    set
    {
      base.buffer = value;
    }
  }

  /**
   * The text font
   */
  public string font
  {
    set
    {
      Pango.FontDescription desc;

      desc =  Pango.FontDescription.from_string (value);
      this.modify_font (desc);
    }
  }

  /**
   * The style scheme
   */
  public string style_scheme
  {
    set
    {
      Gtk.SourceStyleScheme scheme;
      Gtk.SourceStyleSchemeManager manager;

      manager = new Gtk.SourceStyleSchemeManager ();
      if (value != "")
      {
        scheme = manager.get_scheme (value);
      }
      else
      
      {
        scheme = manager.get_scheme ("classic");
      }
      this.buffer.set_style_scheme (scheme);
    }
  }

  private new void scroll_to_iter (Gtk.TextIter iter)
  {
    Gtk.TextMark mark;

    mark = this.buffer.create_mark ("cursor", iter, true);
    this.scroll_to_mark (mark, 0.4, false, 0, 0);
    this.buffer.delete_mark (mark);
  }

  [CCode (instance_pos = -1)]
  private void on_action_goto_line (SourceView sender)
  {
    this.goto_line ();
  }

  [CCode (instance_pos = -1)]
  private void on_action_search (SourceView sender)
  {
    this.search ();
  }

  [CCode (instance_pos = -1)]
  private void on_action_find_next (SourceView sender)
  {
    this.find_next ();
  }

  [CCode (instance_pos = -1)]
  private void on_action_find_prev (SourceView sender)
  {
    this.find_prev ();
  }

  [CCode (instance_pos = -1)]
  private void on_action_replace (SourceView sender)
  {
    this.replace ();
  }

  [CCode (instance_pos = -1)]
  private void on_action_lower (SourceView sender)
  {
    string text;
    Gtk.TextIter end;
    Gtk.TextIter start;

    this.buffer.get_selection_bounds (out start, out end);
    text = this.buffer.get_text (start, end, false);
    text = text.down ();
    this.buffer.delete (start, end);
    this.buffer.insert (start, text, -1);
  }

  [CCode (instance_pos = -1)]
  private void on_action_upper (SourceView sender)
  {
    string text;
    Gtk.TextIter end;
    Gtk.TextIter start;

    this.buffer.get_selection_bounds (out start, out end);
    text = this.buffer.get_text (start, end, false);
    text = text.up ();
    this.buffer.delete (start, end);
    this.buffer.insert (start, text, -1);
  }

  /**
   * Create a new Valide.SourceView
   *
   * @param language The language for coloration
   */
  public SourceView (string language)
  {
    Object (language: language);
  }

  static construct
  {
    unowned Gtk.BindingSet binding_set;

    binding_set = Gtk.BindingSet.by_class (typeof (SourceView).class_ref ());
    action_new (typeof (SourceView), "action_goto_line");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.i, 
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_goto_line", 0);

    action_new (typeof (SourceView), "action_search");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.f,
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_search", 0);

    action_new (typeof (SourceView), "action_find_next");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.g,
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_find_next", 0);

    action_new (typeof (SourceView), "action_find_prev");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.g,
                                 Gdk.ModifierType.CONTROL_MASK
                                   | Gdk.ModifierType.SHIFT_MASK,
                                 "action_find_prev", 0);

    action_new (typeof (SourceView), "action_replace");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.h,
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_replace", 0);

    action_new (typeof (SourceView), "action_lower");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.l,
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_lower", 0);

    action_new (typeof (SourceView), "action_upper");
    Gtk.BindingEntry.add_signal (binding_set, Gdk.KeySyms.u,
                                 Gdk.ModifierType.CONTROL_MASK,
                                 "action_upper", 0);
  }

  construct
  {
    ConfigManager config;
    Gtk.SourceLanguageManager lm;

    lm = new Gtk.SourceLanguageManager ();
    if (this.language != "txt")
    {
      SourceBuffer buffer;

      buffer = new SourceBuffer ();
      buffer.language = lm.get_language (this.language);
      buffer.set_highlight_syntax (true);
      this.buffer = buffer;
    }
    else
    {
      this.buffer = new SourceBuffer ();
    }

    config = ConfigManager.get_instance ();
    this.set_right_margin_position (config.get_integer ("GtkSourceView", "right-margin-position"));
    this.set_show_right_margin (config.get_boolean ("GtkSourceView", "show-right-margin"));
    this.set_show_line_numbers (config.get_boolean ("GtkSourceView", "show-line-numbers"));
    this.set_auto_indent (config.get_boolean ("GtkSourceView", "auto-indent"));
    this.set_highlight_current_line (config.get_boolean ("GtkSourceView", "highlight-current-line"));
    this.set_indent_on_tab (config.get_boolean ("GtkSourceView", "indent-on-tab"));
    this.set_insert_spaces_instead_of_tabs (config.get_boolean ("GtkSourceView", "insert-spaces-instead-of-tabs"));
    this.set_smart_home_end (Gtk.SourceSmartHomeEndType.BEFORE);
    this.set_tab_width (config.get_integer ("GtkSourceView", "tab-width"));
    this.font = config.get_string ("GtkSourceView", "font");
    this.style_scheme = config.get_string ("GtkSourceView", "style-scheme");

    Signal.connect (this, "action_goto_line",
                    (Callback)this.on_action_goto_line, this);
    Signal.connect (this, "action_search",
                    (Callback)this.on_action_search, this);
    Signal.connect (this, "action_find_next",
                    (Callback)this.on_action_find_next, this);
    Signal.connect (this, "action_find_prev",
                    (Callback)this.on_action_find_prev, this);
    Signal.connect (this, "action_replace",
                    (Callback)this.on_action_replace, this);
    Signal.connect (this, "action_lower",
                    (Callback)this.on_action_lower, this);
    Signal.connect (this, "action_upper",
                    (Callback)this.on_action_upper, this);
  }

  /**
   * Scroll text view to cursor
   */
  public void scroll_to_cursor ()
  {
    Gtk.TextBuffer buffer;

    buffer = this.buffer;
    this.scroll_to_mark (buffer.get_insert (), 0.25, false, 0.0, 0.0);
  }

  public bool goto_line (int line = -1)
  {
    bool ret = false;
	  uint line_count;
	  Gtk.TextIter iter;

	  if (line < 0)
	  {
	    if (this.goto_dialog == null)
	    {
        Gtk.Window toplevel;

        this.goto_dialog = new GotoDialog (this);
        toplevel = this.get_ancestor (typeof (Gtk.Window)) as Gtk.Window;
        if (toplevel != null)
        {
          Gtk.WindowGroup group;

          group = toplevel.group;
          if (group != null)
          {
            group.add_window (this.goto_dialog);
          }
        }
      }
	    this.goto_dialog.show ();
	  }
    else
	  {
    	line_count = this.buffer.get_line_count ();

    	if (line <= line_count)
	    {
	      this.grab_focus ();
	    	this.buffer.get_iter_at_line (out iter, line);
      	this.buffer.place_cursor (iter);
        this.scroll_to_iter (iter);
	    	ret = true;
	    }
    }
    return ret;
  }
}

