/* completion.vala
 *
 * Copyright (C) 2008-2010 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

using Valide;

public class Completion : Plugin, Object
{
  private uint ui_id;
  private SourceBuffer buffer;
  private Gtk.ActionGroup action_group = null;

  const Gtk.ActionEntry[] entries = {
    { "search-goto-definition", null, "Go to Definition", "<ctrl>D",
      "Goto the definition of the current symbol", on_goto_definition },
    { "search-autocomplete", null, "_AutoComplete", "<ctrl>space",
      "Display method or symbol information", on_autocomplete }
  };

  const string ui = """
      <ui>
        <menubar name="menubar">
          <menu action="search">
            <placeholder name="search-ops-3">
              <menuitem action="search-goto-definition"/>
              <separator/>
              <menuitem action="search-autocomplete"/>
            </placeholder>
          </menu>
        </menubar>
      </ui>""";

  /**
   * @see Valide.Plugin.path
   */
  public string path { get; construct set; }
  /**
   * @see Valide.Plugin.window
   */
  public Window window { get; construct set; }

  private void get_buffer_info (SourceBuffer buffer, out int pos,
                                out string contents)
  {
    Gtk.TextIter iter;

    contents = buffer.get_buffer_contents ();
    this.buffer.get_iter_at_mark (out iter, this.buffer.get_insert ());
    pos = iter.get_offset ();
  }

  private void jump_to_symbol_definition ()
  {
    int pos;
    string source;
    Valencia.ScanInfo info;

    this.get_buffer_info (this.buffer, out pos, out source);
    info = Valencia.ScanInfo.get_scan_info (this.buffer, false);
    if (info != null && info.inner () != null)
    {
      Valencia.Symbol? sym;
      Valencia.SourceFile sf;
      Valencia.Program program;

      program = Valencia.Program.find_containing (this.buffer.path);
      sf = program.find_source (this.buffer.path);
      sym = sf.resolve (info.inner (), pos, false);
      if (sym != null)
      {
        try
        {
          Document document;

          document = this.window.documents.create (sym.source.filename);
          document.select_offsets (sym.start, sym.start + sym.name_length ());
        }
        catch (Error e)
        {
          warning (e.message);
        }
      }
    }
  }

  private void on_goto_definition ()
  {
    string filename;
    Document document;
    SourceView source_view;

    document = this.window.documents.current;
    source_view = document.text_view;
    filename = source_view.buffer.path;

    if (filename != null && Valencia.Program.is_vala (filename))
    {
      Valencia.Program program;

      program = Valencia.Program.find_containing (filename, true);
      this.buffer = document.buffer;
      if (program.is_parsing ())
      {
        program.system_parse_complete.connect (this.jump_to_symbol_definition);
      }
      else
      {
        this.jump_to_symbol_definition ();
      }
    }
  }

  private void on_autocomplete ()
  {
    Gtk.bindings_activate (this.window.get_focus (), Gdk.KeySyms.space,
                           Gdk.ModifierType.CONTROL_MASK);
  }

  private void setup_ui (DocumentManager sender, Document? document)
  {
    bool active = false;

    if (this.action_group == null)
    {
      Gtk.UIManager ui_manager;

      ui_manager = this.window.ui_manager;
      this.action_group = new Gtk.ActionGroup ("completion");
      this.action_group.add_actions (this.entries, this);
      ui_manager.insert_action_group (this.action_group, 0);
      try
      {
        this.ui_id = ui_manager.add_ui_from_string (ui, -1);
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }

    if (sender.current != null)
    {
      active = true;
    }
    this.action_group.get_action ("search-goto-definition").sensitive = active;
    this.action_group.get_action ("search-autocomplete").sensitive = active;
  }

  private void on_tab_added (DocumentManager sender, Document document)
  {
    try
    {
      string filename;
      Gtk.SourceView source_view;
      Gtk.SourceCompletion completion;

      filename = document.path;
      source_view = document.split_view.active_view;
      completion = source_view.get_completion ();
      completion.show_headers = false;
      completion.remember_info_visibility = true;
      completion.add_provider ((Gtk.SourceCompletionProvider)new ValenciaProvider (document.buffer));

      this.setup_ui (this.window.documents, null);
    }
    catch (Error e)
    {
      debug (e.message);
    }
  }

  construct
  {
    this.window.documents.tab_added.connect (this.on_tab_added);
    this.window.documents.tab_removed.connect (this.setup_ui);
    this.setup_ui (this.window.documents, null);
  }

  ~Completion ()
  {
    this.window.ui_manager.remove_ui (this.ui_id);
  }
}

[ModuleInit]
public Type register_plugin (void* user_data) throws PluginError
{
  return typeof (Completion);
}

