/* wizard.vala
 *
 * Copyright (C) 2010  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

using Valide;

class WizardPlugin : Plugin, Object
{
  private uint ui_id;
  private Gtk.ActionGroup action_group = null;

  const Gtk.ActionEntry[] entries = {
    { "project-wizard-class", null, "New class", "", "Create a new class", on_create },
    { "project-wizard-interface", null, "New Interface", "", "Create a new interface", on_create },
    { "project-wizard-enum", null, "New enumeration", "", "Create a new enumeration", on_create },
    { "project-wizard-struct", null, "New structure", "", "Create a vala structure", on_create }
  };

  const string ui = """
      <ui>
        <menubar name="menubar">
          <menu action="project">
            <placeholder name="project-ops-1">
              <separator/>
              <menuitem action="project-wizard-class"/>
              <menuitem action="project-wizard-interface"/>
              <menuitem action="project-wizard-enum"/>
              <menuitem action="project-wizard-struct"/>
            </placeholder>
          </menu>
        </menubar>
      </ui>""";

  /**
   * @see Plugin.path
   */
  public string path { get; /*construct*/ set; }
  /**
   * @see Plugin.window
   */
  public Window window { get; /*construct*/ set; }

  /**
   * Generates a Template and flushs it to the harddisk
   */
  private void generate(string classname, string folder, string baseclass, string[] interfaces, bool header, bool main)
  {
    string newfile = Path.build_filename(folder, classname + ".vala");

    var t = new FileTemplate("public",Type,classname, baseclass, interfaces, header, main);
    t.flushto(newfile);
  }

  private string[] splitInterfaces(string ifs)
  {
    if (ifs == "")
    {
      return new string[0];
    }
    else if (!ifs.contains(","))
    {
      return new string[] { ifs };
    }
    return ifs.split(",");
  }

  private void on_create (Gtk.Action action)
  {
    WizardInfo info;
    WizardDialog dialog;

    info = WizardInfo ();
    info.title = action.label;
    info.type = action.name.substring (15 /* "project-wizard-".lenght() */);
    //info.path = Utils.getFolder(RealPath);

    dialog = new WizardDialog (info);
    dialog.project_manager = this.window.projects;
    dialog.show_all ();
    // pressed ok -> returns 0
    if (dialog.run () == 0)
    {
      string folder = this.widgets.folder_entry.get_text();
      string classname = this.widgets.class_entry.get_text();
      string baseclass = this.widgets.super_entry.get_text();
      string interfaces = this.widgets.interface_entry.get_text();
      bool header = this.widgets.template_header.get_active();
      bool main = this.widgets.addmain.get_active();

      string[] ifs = splitInterfaces(interfaces);

      generate(classname, folder, baseclass, ifs, header, main);

      //TODO add the new file to the PorjectManager to enable code completition
      //project_manager.refresh();
    }
    dialog.destroy ();
  }

  private void setup_ui ()
  {
    bool active = false;

    if (this.action_group == null)
    {
      Gtk.UIManager ui_manager;

      ui_manager = this.window.ui_manager;
      this.action_group = new Gtk.ActionGroup ("wizard-plugin");
      this.action_group.add_actions (this.entries, this);
      ui_manager.insert_action_group (this.action_group, 0);
      try
      {
        this.ui_id = ui_manager.add_ui_from_string (this.ui, -1);
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }

    if ((this.window.documents.current != null
          && this.window.documents.current is AbstractTextDocument)
        || this.window.projects.current != null)
    {
      active = true;
    }
    //this.action_group.get_action ("search-advanced").sensitive = active;
  }

  construct
  {
    this.setup_ui ();
  }

  ~WizardPlugin ()
  {
    this.window.ui_manager.remove_ui (this.ui_id);
  }
}

[ModuleInit]
public Type register_plugin (void* user_data) throws PluginError
{
  return typeof (WizardPlugin);
}

